@@ -31,7 +31,7 @@ import datetime
 import time
 import collections
 
-VERSION = (0, 3, 5)
+VERSION = (0, 3, 6)
 __version__ = '%s.%s.%s' % VERSION
 apilevel = '2.0'
 threadsafety = 1
@@ -174,30 +174,30 @@ PG_TYPE_ANYENUM = 3500
 PG_TYPE_FDW_HANDLER = 3115
 PG_TYPE_ANYRANGE = 3831
 
-def _decode_column(data, oid, encoding):
-    class TZ(datetime.tzinfo):
-        def __init__(self, offset):
-            self.offset = offset
-            hours = int(self.offset[:3])
-            if len(offset) > 3:
-                minutes = int(self.offset[4:6])
-            else:
-                minutes = 0
-            if len(offset) > 6:
-                seconds = int(self.offset[7:])
-            else:
-                seconds = 0
-            self.delta = datetime.timedelta(hours=hours, minutes=minutes, seconds=seconds)
+class TZ(datetime.tzinfo):
+    def __init__(self, offset):
+        self.offset = offset
+        hours = int(self.offset[:3])
+        if len(offset) > 3:
+            minutes = int(self.offset[4:6])
+        else:
+            minutes = 0
+        if len(offset) > 6:
+            seconds = int(self.offset[7:])
+        else:
+            seconds = 0
+        self.delta = datetime.timedelta(hours=hours, minutes=minutes, seconds=seconds)
 
-        def utcoffset(self, dt):
-            return self.delta
+    def utcoffset(self, dt):
+        return self.delta
 
-        def dst(self, dt):
-            return self.delta
+    def dst(self, dt):
+        return self.delta
 
-        def tzname(self, dt):
-            return u"UTC" + self.offset
+    def tzname(self, dt):
+        return u"UTC" + self.offset
 
+def _decode_column(data, oid, encoding, tzinfo, use_tzinfo):
     if data is None:
         return data
     data = data.decode(encoding)
@@ -217,12 +217,18 @@ def _decode_column(data, oid, encoding):
             dt = datetime.datetime.strptime(data, '%H:%M:%S')
         else:
             dt = datetime.datetime.strptime(data, '%H:%M:%S.%f')
-        return datetime.time(dt.hour, dt.minute, dt.second, dt.microsecond)
+        dt = datetime.time(dt.hour, dt.minute, dt.second, dt.microsecond)
+        if tzinfo:
+            dt = dt.replace(tzinfo=tzinfo)
+        return dt
     elif oid in (PG_TYPE_TIMESTAMP, ):
         if len(data) == 19:
-            return datetime.datetime.strptime(data, '%Y-%m-%d %H:%M:%S')
+            dt = datetime.datetime.strptime(data, '%Y-%m-%d %H:%M:%S')
         else:
-            return datetime.datetime.strptime(data, '%Y-%m-%d %H:%M:%S.%f')
+            dt = datetime.datetime.strptime(data, '%Y-%m-%d %H:%M:%S.%f')
+        if tzinfo:
+            dt = dt.replace(tzinfo=tzinfo)
+        return dt
     elif oid in (PG_TYPE_TIMETZ, ):
         n = data.rfind('+')
         if n == -1:
@@ -233,7 +239,8 @@ def _decode_column(data, oid, encoding):
             dt = datetime.datetime.strptime(s, '%H:%M:%S')
         else:
             dt = datetime.datetime.strptime(s, '%H:%M:%S.%f')
-        dt.replace(tzinfo=TZ(data[n:]))
+        if use_tzinfo:
+            dt = dt.replace(tzinfo=TZ(data[n:]))
         return datetime.time(dt.hour, dt.minute, dt.second, dt.microsecond, tzinfo=dt.tzinfo)
     elif oid in (PG_TYPE_TIMESTAMPTZ, ):
         n = data.rfind('+')
@@ -244,7 +251,9 @@ def _decode_column(data, oid, encoding):
             dt = datetime.datetime.strptime(s, '%Y-%m-%d %H:%M:%S')
         else:
             dt = datetime.datetime.strptime(s, '%Y-%m-%d %H:%M:%S.%f')
-        return dt.replace(tzinfo=TZ(data[n:]))
+        if use_tzinfo:
+            dt = dt.replace(tzinfo=TZ(data[n:]))
+        return dt
     elif oid in (PG_TYPE_INTERVAL, ):
         dt = data.split('days')
         if len(dt) < 2:
@@ -298,47 +307,6 @@ def _bint_to_bytes(val):    # Convert int value to big endian 4 bytes.
     else:
         return bytes([(val >> 24) & 0xff, (val >> 16) & 0xff, (val >> 8) & 0xff, val & 0xff])
 
-_escape_hook = {}
-
-def add_escape_hook(t, func):
-    _escape_hook[t] = func
-
-def escape_parameter(v):
-    t = type(v)
-    func = _escape_hook.get(t)
-    if func:
-        return func(v)
-    if v is None:
-        return 'NULL'
-    elif (PY2 and t == unicode) or (not PY2 and t == str):  # string
-        return u"'" + v.replace(u"'", u"''") + u"'"
-    elif PY2 and t == str:    # PY2 str
-        v = ''.join(['\\%03o' % (ord(c), ) if ord(c) < 32 or ord(c) > 127 else c for c in v])
-        return "'" + v.replace("'", "''") + "'"
-    elif t == bytearray or t == bytes:        # binary
-        return "'" + ''.join(['\\%03o' % (c, ) for c in v]) + "'::bytea"
-    elif t == bool:
-        return u"TRUE" if v else u"FALSE"
-    elif t == time.struct_time:
-        return u'%04d-%02d-%02d %02d:%02d:%02d' % (
-            v.tm_year, v.tm_mon, v.tm_mday, v.tm_hour, v.tm_min, v.tm_sec)
-    elif t == datetime.datetime:
-        return "'" + str(v) + "'"
-    elif t == datetime.date:
-        return "'" + str(v) + "'"
-    elif t == datetime.timedelta:
-        if v.seconds:
-            return u"interval '%d second'" % (v.days * 86400 + v.seconds, )
-        else:
-            return u"interval '%d day'" % (v.days, )
-    elif t == int or t == float or (PY2 and t == long):
-        return str(v)
-    elif t == decimal.Decimal:
-        return "'" + str(v) + "'"
-    elif t == list or t == tuple:
-        return u'{' + u','.join([escape_parameter(e) for e in v]) + u'}'
-    else:
-        return "'" + str(v) + "'"
 
 Date = datetime.date
 Time = datetime.time
@@ -428,7 +396,6 @@ class Cursor(object):
         pass
 
     def execute(self, query, args=()):
-        if DEBUG: DEBUG_OUTPUT('Cursor::execute()\t%s\t%s' % (query, args))
         if not self.connection or not self.connection.is_connect():
             raise ProgrammingError(u"08003:Lost connection")
         self.description = []
@@ -436,14 +403,13 @@ class Cursor(object):
         self.query = query
         self.args = args
         if args:
-            escaped_args = tuple(escape_parameter(arg) for arg in args)
+            escaped_args = tuple(self.connection.escape_parameter(arg) for arg in args)
             query = query.replace('%', '%%').replace('%%s', '%s')
             query = query % escaped_args
             query = query.replace('%%', '%')
         self.connection.execute(query, self)
 
     def executemany(self, query, seq_of_params):
-        if DEBUG: DEBUG_OUTPUT('Cursor::executemany()\t%s\t%s' % (query, seq_of_params))
         rowcount = 0
         for params in seq_of_params:
             self.execute(query, params)
@@ -451,7 +417,6 @@ class Cursor(object):
         self._rowcount = rowcount
 
     def fetchone(self):
-        if DEBUG: DEBUG_OUTPUT('Cursor::fetchone()')
         if not self.connection or not self.connection.is_connect():
             raise OperationalError(u"08003:Lost connection")
         if len(self._rows):
@@ -483,6 +448,7 @@ class Cursor(object):
     def closed(self):
         return self.connection is None or not self.connection.is_connect()
 
+
     def __iter__(self):
         return self
 
@@ -496,19 +462,21 @@ class Cursor(object):
         return self.__next__()
 
 class Connection(object):
-    def __init__(self, user, password, database, host, port, timeout, use_ssl):
-        if DEBUG: DEBUG_OUTPUT("Connection::__init__()")
+    def __init__(self, user, password, database, host, port, tzinfo, timeout, use_ssl):
         self.user = user
         self.password = password
         self.database = database
         self.host = host
         self.port = port
+        self.tzinfo = tzinfo
         self.timeout = timeout
         self.use_ssl = use_ssl
         self.encoding = 'UTF8'
         self.autocommit = False
-        self.in_transaction = False
+        self._ready_for_query = b'I'
+        self.use_tzinfo = True
         self._open()
+        self.encoders = {}
 
     def __enter__(self):
         return self
@@ -535,7 +503,7 @@ class Connection(object):
             data = self._read(ln)
             if code == PG_B_READY_FOR_QUERY:
                 if DEBUG: DEBUG_OUTPUT("READY_FOR_QUERY:%s" % (data, ))
-                self.in_transaction = (data == b'T')
+                self._ready_for_query = data
                 break
             elif code == PG_B_AUTHENTICATION:
                 auth_method = _bytes_to_bint(data[:4])
@@ -610,14 +578,14 @@ class Connection(object):
                             row.append(data[n:n+ln])
                             n += ln
                     for i in range(len(row)):
-                        row[i] = _decode_column(row[i], obj.description[i][1], self.encoding)
+                        row[i] = _decode_column(row[i], obj.description[i][1], self.encoding, self.tzinfo, self.use_tzinfo)
                     obj._rows.append(tuple(row))
                     if DEBUG: DEBUG_OUTPUT("\t\t%s" % (row, ))
             elif code == PG_B_NOTICE_RESPONSE:
                 if DEBUG: DEBUG_OUTPUT("NOTICE_RESPONSE:%s" % (HEX(data), ))
                 for s in data.split(b'\x00'):
                     if DEBUG: DEBUG_OUTPUT("\t\t%s" % (s.decode(self.encoding)))
-            elif code == PG_B_ERROR_RESPONSE:
+            elif code == PG_B_ERROR_RESPONSE and not errobj:
                 if DEBUG: DEBUG_OUTPUT("ERROR_RESPONSE:%s" % (HEX(data), ))
                 err = data.split(b'\x00')
                 for b in err:
@@ -628,7 +596,7 @@ class Connection(object):
                 if not PY2:
                     message = message.decode(self.encoding)
                 if errcode[:2] == b'23':
-                    errobj = IntegrityError( message)
+                    errobj = IntegrityError(message)
                 else:
                     errobj = DatabaseError(message)
             elif code == PG_B_COPY_OUT_RESPONSE:
@@ -651,9 +619,7 @@ class Connection(object):
                 self._write(b'c\x00\x00\x00\x04S\x00\x00\x00\x04')
             else:
                 if DEBUG: DEBUG_OUTPUT("SKIP:%d\t%i\t%s" % (code, ln, HEX(data)))
-        if errobj:
-            raise errobj
-        return
+        return errobj
 
     def _read(self, ln):
         if not self.sock:
@@ -698,7 +664,50 @@ class Connection(object):
         v += b'\x00'
 
         self._write(_bint_to_bytes(len(v) + 4) + v)
-        self._process_messages(None)
+        err = self._process_messages(None)
+        if err:
+            raise err
+
+    def escape_parameter(self, v):
+        t = type(v)
+        func = self.encoders.get(t)
+        if func:
+            return func(v)
+        if v is None:
+            return 'NULL'
+        elif (PY2 and t == unicode) or (not PY2 and t == str):  # string
+            return u"'" + v.replace(u"'", u"''") + u"'"
+        elif PY2 and t == str:    # PY2 str
+            v = ''.join(['\\%03o' % (ord(c), ) if ord(c) < 32 or ord(c) > 127 else c for c in v])
+            return "'" + v.replace("'", "''") + "'"
+        elif t == bytearray or t == bytes:        # binary
+            return "'" + ''.join(['\\%03o' % (c, ) for c in v]) + "'::bytea"
+        elif t == bool:
+            return u"TRUE" if v else u"FALSE"
+        elif t == time.struct_time:
+            return u'%04d-%02d-%02d %02d:%02d:%02d' % (
+                v.tm_year, v.tm_mon, v.tm_mday, v.tm_hour, v.tm_min, v.tm_sec)
+        elif t == datetime.datetime:
+            return "'" + str(v) + "'"
+        elif t == datetime.date:
+            return "'" + str(v) + "'"
+        elif t == datetime.timedelta:
+            if v.seconds:
+                return u"interval '%d second'" % (v.days * 86400 + v.seconds, )
+            else:
+                return u"interval '%d day'" % (v.days, )
+        elif t == int or t == float or (PY2 and t == long):
+            return str(v)
+        elif t == decimal.Decimal:
+            return "'" + str(v) + "'"
+        elif t == list or t == tuple:
+            return u'{' + u','.join([self.escape_parameter(e) for e in v]) + u'}'
+        else:
+            return "'" + str(v) + "'"
+
+    @property
+    def is_dirty(self):
+        return self._ready_for_query in b'TE'
 
     def is_connect(self):
         return bool(self.sock)
@@ -707,17 +716,19 @@ class Connection(object):
         return Cursor(self)
 
     def _execute(self, query, obj=None):
+        if DEBUG: DEBUG_OUTPUT('Connection::_execute()\t%s' % (query, ))
         self._send_message(
             PG_F_QUERY,
             query.encode(self.encoding) + b'\x00',
         )
-        self._process_messages(obj)
+        err = self._process_messages(obj)
+        if err:
+            raise err
         if self.autocommit:
             self.commit()
 
     def execute(self, query, obj=None):
-        if DEBUG: DEBUG_OUTPUT('Connection::execute()\t%s' % (query, ))
-        if not self.in_transaction:
+        if self._ready_for_query != b'T':
             self.begin()
         self._execute(query, obj)
 
@@ -726,6 +737,8 @@ class Connection(object):
 
     def begin(self):
         if DEBUG: DEBUG_OUTPUT('BEGIN')
+        if self._ready_for_query == b'E':
+            self.rollback()
         self._send_message(PG_F_QUERY, b"BEGIN\x00")
         self._process_messages(None)
 
@@ -735,11 +748,17 @@ class Connection(object):
             self._send_message(PG_F_QUERY, b"COMMIT\x00")
             self._process_messages(None)
 
-    def rollback(self):
-        if DEBUG: DEBUG_OUTPUT('ROLLBACK')
+    def _rollback(self):
         if self.sock:
+            if DEBUG: DEBUG_OUTPUT('ROLLBACK')
             self._send_message(PG_F_QUERY, b"ROLLBACK\x00")
-            self._process_messages(None)
+            return self._process_messages(None)
+        return None
+
+    def rollback(self):
+        err = self._rollback()
+        if err:
+            raise err
 
     def reopen(self):
         self.close()
@@ -753,6 +772,6 @@ class Connection(object):
             self.sock.close()
             self.sock = None
 
-def connect(host, user, password='', database=None, port=5432, timeout=None, use_ssl=False):
-    return Connection(user, password, database, host, port, timeout, use_ssl)
+def connect(host, user, password='', database=None, port=5432, tzinfo=None, timeout=None, use_ssl=False):
+    return Connection(user, password, database, host, port, tzinfo, timeout, use_ssl)
 
